package net.knarcraft.blockhunt;

import me.libraryaddict.disguise.DisguiseAPI;
import me.libraryaddict.disguise.disguisetypes.DisguiseType;
import me.libraryaddict.disguise.disguisetypes.MiscDisguise;
import net.knarcraft.blockhunt.arena.Arena;
import net.knarcraft.blockhunt.arena.ArenaState;
import net.knarcraft.blockhunt.config.MessageKey;
import net.knarcraft.blockhunt.listener.OnBlockBreakEvent;
import net.knarcraft.blockhunt.listener.OnBlockPlaceEvent;
import net.knarcraft.blockhunt.listener.OnEntityDamageByEntityEvent;
import net.knarcraft.blockhunt.listener.OnEntityDamageEvent;
import net.knarcraft.blockhunt.listener.OnFoodLevelChangeEvent;
import net.knarcraft.blockhunt.listener.OnInventoryClickEvent;
import net.knarcraft.blockhunt.listener.OnInventoryCloseEvent;
import net.knarcraft.blockhunt.listener.OnPlayerDropItemEvent;
import net.knarcraft.blockhunt.listener.OnPlayerInteractEvent;
import net.knarcraft.blockhunt.listener.OnPlayerMoveEvent;
import net.knarcraft.blockhunt.listener.OnPlayerQuitEvent;
import net.knarcraft.blockhunt.listener.OnPlayerSwapHandItemsEvent;
import net.knarcraft.blockhunt.listener.OnPlayerTeleportEvent;
import net.knarcraft.blockhunt.listener.OnSignChangeEvent;
import net.knarcraft.blockhunt.manager.CommandManager;
import net.knarcraft.blockhunt.manager.ConfigManager;
import net.knarcraft.blockhunt.util.ArenaHelper;
import net.knarcraft.blockhunt.util.MessageHelper;
import net.knarcraft.blockhunt.util.PlayerHelper;
import net.knarcraft.blockhunt.util.ScoreboardHelper;
import net.knarcraft.blockhunt.util.SignsHelper;
import net.knarcraft.blockhunt.util.VisibilityHelper;
import org.bukkit.Bukkit;
import org.bukkit.GameMode;
import org.bukkit.Location;
import org.bukkit.Material;
import org.bukkit.Sound;
import org.bukkit.World;
import org.bukkit.configuration.serialization.ConfigurationSerialization;
import org.bukkit.enchantments.Enchantment;
import org.bukkit.entity.Player;
import org.bukkit.event.Listener;
import org.bukkit.inventory.ItemStack;
import org.bukkit.plugin.PluginDescriptionFile;
import org.bukkit.plugin.java.JavaPlugin;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import static net.knarcraft.blockhunt.util.StringHelper.uppercaseAllFirst;

/**
 * BlockHunt's main class
 */
public class BlockHunt extends JavaPlugin implements Listener {

    private static PluginDescriptionFile pluginDescriptionFile;
    public static BlockHunt plugin;
    public String rootPermission;

    @Override
    public void onEnable() {
        getServer().getPluginManager().registerEvents(this, this);

        getServer().getPluginManager().registerEvents(new OnBlockBreakEvent(), this);
        getServer().getPluginManager().registerEvents(new OnBlockPlaceEvent(), this);
        getServer().getPluginManager().registerEvents(new OnEntityDamageByEntityEvent(), this);
        getServer().getPluginManager().registerEvents(new OnEntityDamageEvent(), this);
        getServer().getPluginManager().registerEvents(new OnFoodLevelChangeEvent(), this);
        getServer().getPluginManager().registerEvents(new OnInventoryClickEvent(), this);
        getServer().getPluginManager().registerEvents(new OnInventoryCloseEvent(), this);
        getServer().getPluginManager().registerEvents(new OnPlayerSwapHandItemsEvent(), this);

        getServer().getPluginManager().registerEvents(new OnPlayerDropItemEvent(), this);
        getServer().getPluginManager().registerEvents(new OnPlayerInteractEvent(), this);
        getServer().getPluginManager().registerEvents(new OnPlayerMoveEvent(), this);
        getServer().getPluginManager().registerEvents(new OnPlayerTeleportEvent(), this);
        getServer().getPluginManager().registerEvents(new OnPlayerQuitEvent(), this);
        getServer().getPluginManager().registerEvents(new OnSignChangeEvent(), this);

        ConfigurationSerialization.registerClass(Arena.class, "BlockHuntArena");

        BlockHunt.pluginDescriptionFile = getDescription();
        this.rootPermission = BlockHunt.getPluginDescriptionFile().getName().toLowerCase() + ".";
        plugin = this;

        ConfigManager.newFiles();

        CommandManager.load();

        if (!getServer().getPluginManager().isPluginEnabled("LibsDisguises")) {
            MessageHelper.broadcastMessage(MessageKey.ERROR_LIBS_DISGUISES_NOT_INSTALLED);
            Bukkit.getServer().getPluginManager().disablePlugin(this);
            return;
        }

        if (!getServer().getPluginManager().isPluginEnabled("ProtocolLib")) {
            MessageHelper.broadcastMessage(MessageKey.ERROR_PROTOCOL_LIB_NOT_INSTALLED);
            Bukkit.getServer().getPluginManager().disablePlugin(this);
            return;
        }

        ArenaHelper.loadArenas();

        MessageHelper.sendMessage(null, MessageKey.LOG_ENABLED_PLUGIN, "name-" +
                        BlockHunt.pluginDescriptionFile.getName(), "version-" + BlockHunt.pluginDescriptionFile.getVersion(),
                "authors-" + BlockHunt.pluginDescriptionFile.getAuthors().get(0));

        // Welcome to the massive game loop!!
        getServer().getScheduler().runTaskTimer(this, this::updateArenaStates, 0, 20);
    }

    @Override
    public void onDisable() {
        for (Arena arena : MemoryStorage.arenaMap.values()) {
            String cause = "[BlockHunt] Arena " + arena.arenaName + " has been stopped";
            ArenaHelper.stopArena(arena, cause, MessageKey.WARNING_ARENA_STOPPED);
        }

        MessageHelper.sendMessage(null, MessageKey.LOG_DISABLED_PLUGIN, "name-" +
                        BlockHunt.pluginDescriptionFile.getName(), "version-" + BlockHunt.pluginDescriptionFile.getVersion(),
                "authors-" + BlockHunt.pluginDescriptionFile.getAuthors().get(0));
    }

    public static PluginDescriptionFile getPluginDescriptionFile() {
        return pluginDescriptionFile;
    }

    /**
     * Shorten a String for things like the Scoreboard title.
     *
     * @param string    <p>The string to be shortened</p>
     * @param maxLength <p>The maximum amount of characters allowed</p>
     * @return <p>The shortened string, or the original string if already short enough</p>
     */
    public static String cutString(String string, int maxLength) {
        if (string.length() <= maxLength) {
            return string;
        } else {
            return string.substring(0, maxLength);
        }
    }

    private void updateArenaStates() {
        for (Arena arena : MemoryStorage.arenaMap.values()) {
            if (arena.gameState == ArenaState.WAITING) {
                if (arena.playersInArena.size() >= arena.minPlayers) {
                    arena.gameState = ArenaState.STARTING;
                    arena.timer = arena.timeInLobbyUntilStart;
                    ArenaHelper.sendMessage(arena, MessageKey.NORMAL_LOBBY_ARENA_IS_STARTING, "1-" +
                            arena.timeInLobbyUntilStart);
                }
            } else if (arena.gameState == ArenaState.STARTING) {
                updateStartingArena(arena);
            }

            // TODO: Figure out a better way to do this. Currently, individual seeker timers are used to allow another
            //  player to take over if a seeker leaves the game. But assuming the game is already running, wouldn't it
            //  make more sense to allow the seeker to seek immediately?
            for (Player player : arena.seekers) {
                // Reduce the seeker wait timer if necessary
                Integer seekerTime = MemoryStorage.seekerTime.get(player);
                if (seekerTime != null) {
                    MemoryStorage.seekerTime.put(player, --seekerTime);
                    if (seekerTime <= 0) {
                        // Teleport the seeker to the hiders warp, and allow them to seek
                        PlayerHelper.teleport(player, arena.arenaLocations.getHidersWarp());
                        MemoryStorage.seekerTime.remove(player);
                        ArenaHelper.sendMessage(arena, MessageKey.NORMAL_IN_GAME_SEEKER_SPAWNED,
                                "%playerName%-" + player.getName());
                    }
                }
            }

            if (arena.gameState == ArenaState.IN_GAME) {
                updateInGameArena(arena);
            }

            // Display countdown as the player's XP level
            for (Player player : arena.playersInArena) {
                player.setLevel(arena.timer);
                if (arena.seekers.contains(player)) {
                    player.setGameMode(GameMode.SURVIVAL);
                } else {
                    player.setGameMode(GameMode.ADVENTURE);
                }
            }
        }
        SignsHelper.updateSigns(); //TODO Only do this when needed (game-state change or player count change)
    }

    /**
     * Updates the state of an arena that's in the process of starting
     *
     * @param arena <p>An arena in the STARTING state</p>
     */
    private void updateStartingArena(Arena arena) {
        // Reduce the arena's timer
        arena.timer -= 1;
        if (arena.timer > 0) {
            if (arena.timer == 60 || arena.timer == 30 || arena.timer == 10) {
                // Announce when the arena starts
                ArenaHelper.sendMessage(arena, MessageKey.NORMAL_LOBBY_ARENA_IS_STARTING, "1-" + arena.timer);
            } else if (arena.timer <= 5) {
                // Count down from 5 seconds
                World world = null;
                Location warp = arena.arenaLocations.getLobbyWarp();
                if (warp != null) {
                    world = warp.getWorld();
                }
                if (world != null) {
                    int pitch = (int) (2 - Math.floor(arena.timer / 2.0));
                    world.playSound(arena.arenaLocations.getLobbyWarp(), Sound.ENTITY_EXPERIENCE_ORB_PICKUP, 1,
                            pitch);
                }
                ArenaHelper.sendMessage(arena, MessageKey.NORMAL_LOBBY_ARENA_IS_STARTING, "1-" + arena.timer);
            }
        } else {
            startArena(arena);
        }
    }

    /**
     * Starts an arena
     *
     * @param arena <p>The arena to start</p>
     */
    private void startArena(Arena arena) {
        // Update the game state, and set the timer to count down from the specified game-time
        arena.gameState = ArenaState.IN_GAME;
        arena.timer = arena.gameTime;

        // Assign the necessary seekers
        assignSeekers(arena);

        // Assign the necessary hiders
        assignHiders(arena);

        MessageHelper.broadcastMessage(arena.hiders, MessageKey.NORMAL_LOBBY_ARENA_STARTED_HIDERS,
                "seconds-" + arena.waitingTimeSeeker);
        MessageHelper.broadcastMessage(arena.seekers, MessageKey.NORMAL_LOBBY_ARENA_STARTED_SEEKERS,
                "seconds-" + arena.waitingTimeSeeker);
    }

    /**
     * Assigns all non-seekers in an arena as hiders
     *
     * @param arena <p>The arena to assign hiders for</p>
     */
    private void assignHiders(Arena arena) {
        List<Player> possibleHiders = new ArrayList<>(arena.playersInArena);
        possibleHiders.removeAll(arena.seekers);

        // Assign all non-seekers to hiders
        for (Player arenaPlayer : possibleHiders) {
            arena.hiders.add(arenaPlayer);
            // Clear the player's inventory
            arenaPlayer.getInventory().clear();

            // Assign a random disguise block
            ItemStack block = arena.disguiseBlocks.get(MemoryStorage.random.nextInt(arena.disguiseBlocks.size()));
            // If the player has chosen their block type, use that instead
            if (MemoryStorage.chosenBlock.get(arenaPlayer) != null) {
                block = MemoryStorage.chosenBlock.get(arenaPlayer);
                MemoryStorage.chosenBlock.remove(arenaPlayer);
            }

            // Disguise the hider from all players
            MiscDisguise disguise = new MiscDisguise(DisguiseType.FALLING_BLOCK, block.getType());
            DisguiseAPI.disguiseToAll(arenaPlayer, disguise);

            // Teleport to the hiders warp
            PlayerHelper.teleport(arenaPlayer, arena.arenaLocations.getHidersWarp());

            // Put the block in the player's inventory and on the player's head, so they know which block they are
            ItemStack blockTimer = new ItemStack(block.getType(), 5);
            arenaPlayer.getInventory().setItem(8, blockTimer);
            arenaPlayer.getInventory().setHelmet(new ItemStack(block));
            MemoryStorage.playerBlock.put(arenaPlayer, block);

            // Tell the player which block they are disguised as
            String blockName = block.getType().name();
            blockName = uppercaseAllFirst(blockName.replace("_", " "));
            MessageHelper.sendMessage(arenaPlayer, MessageKey.NORMAL_IN_GAME_BLOCK,
                    "block-" + blockName);
        }
    }

    /**
     * Assigns necessary seekers for the given arena
     *
     * @param arena <p>The arena to assign seekers to</p>
     */
    private void assignSeekers(Arena arena) {
        List<Player> possibleSeekers = new ArrayList<>(arena.playersInArena);

        // Prioritize players that have asked to be seekers
        int seekers = 0;
        for (Player player : arena.playersInArena) {
            /* Add the player as a seeker if they prefer to be a seeker, or remove them from possible seekers if they
               have set themselves as a hider */
            Boolean chosenSeeker = MemoryStorage.chosenSeeker.get(player);
            if (chosenSeeker != null) {
                if (chosenSeeker) {
                    addSeeker(arena, player);
                }
                possibleSeekers.remove(player);
            }

            // Prevent more than the max amount of seekers from being assigned
            if (seekers >= arena.amountSeekersOnStart) {
                break;
            }
        }

        // Randomly assign the required amount of seekers
        for (int i = seekers; i < arena.amountSeekersOnStart; i++) {
            // Prevent an infinite loop if it's impossible to assign the required amount of seekers
            if (possibleSeekers.size() == 0) {
                break;
            }

            // Add a random seeker
            Player seeker = possibleSeekers.remove(MemoryStorage.random.nextInt(possibleSeekers.size()));
            addSeeker(arena, seeker);
        }
    }

    /**
     * Adds a single seeker to an arena, and teleports them to the seekers warp
     *
     * @param arena  <p>The arena to assign a seeker for</p>
     * @param player <p>The player to assign as a seeker</p>
     */
    private void addSeeker(Arena arena, Player player) {
        ArenaHelper.sendMessage(arena, MessageKey.NORMAL_IN_GAME_SEEKER_CHOSEN, "seeker-" + player.getName());
        arena.seekers.add(player);
        PlayerHelper.teleport(player, arena.arenaLocations.getSeekersWarp());
        player.getInventory().clear();
        player.setWalkSpeed(0.3F);
        MemoryStorage.seekerTime.put(player, arena.waitingTimeSeeker);

        // Give the seeker their armor and sword
        giveSeekerItems(player);
    }

    /**
     * Gives a seeker their armor, sword and bow items
     *
     * @param player <p>The player to give seeker items to</p>
     */
    private void giveSeekerItems(Player player) {
        ItemStack swordItem = new ItemStack(Material.DIAMOND_SWORD, 1);
        swordItem.addUnsafeEnchantment(Enchantment.DAMAGE_ALL, 10);
        player.getInventory().setItem(0, swordItem);

        player.getInventory().setHelmet(new ItemStack(Material.IRON_HELMET, 1));
        player.getInventory().setChestplate(new ItemStack(Material.IRON_CHESTPLATE, 1));
        player.getInventory().setLeggings(new ItemStack(Material.IRON_LEGGINGS, 1));
        player.getInventory().setBoots(new ItemStack(Material.IRON_BOOTS, 1));

        ItemStack infinityBowItem = new ItemStack(Material.BOW, 1);
        infinityBowItem.addEnchantment(Enchantment.ARROW_INFINITE, 1);
        player.getInventory().setItem(1, infinityBowItem);
        player.getInventory().setItem(2, new ItemStack(Material.ARROW, 1));
        player.playSound(player.getLocation(), Sound.BLOCK_ANVIL_USE, 1, 1);
    }

    /**
     * Updates an arena that's currently in-game
     *
     * @param arena <P>The arena to update</P>
     */
    private void updateInGameArena(Arena arena) {
        arena.timer--;

        // If the game ends, trigger the hiders' win
        if (arena.timer <= 0) {
            ArenaHelper.hidersWin(arena);
            return;
        }

        // If "timeUntilHidersSword" has passed, give the hiders their sword
        if (arena.timer == arena.gameTime - arena.timeUntilHidersSword) {
            ItemStack sword = new ItemStack(Material.WOODEN_SWORD, 1);
            sword.addUnsafeEnchantment(Enchantment.KNOCKBACK, 2);

            for (Player arenaPlayer : arena.hiders) {
                arenaPlayer.getInventory().addItem(sword);
                MessageHelper.sendMessage(arenaPlayer, MessageKey.NORMAL_IN_GAME_GIVEN_SWORD);
            }
        }

        // Announce the type of blocks the remaining hiders are disguised as
        if (arena.blockAnnouncerTime > 0 && arena.timer == arena.blockAnnouncerTime) {
            Set<String> remainingBlocks = new HashSet<>();
            for (Player arenaPlayer : arena.hiders) {
                ItemStack item = arenaPlayer.getInventory().getItem(8);
                String block = (item == null ? Material.AIR : item.getType()).name();
                remainingBlocks.add(uppercaseAllFirst(block.replace("_", " ")));
            }
            String blocklist = String.join(", ", remainingBlocks);
            ArenaHelper.sendMessage(arena, MessageKey.NORMAL_IN_GAME_BLOCKS_LEFT, "1-" + blocklist);
        }

        // Count down towards the end of the arena
        if (arena.timer == 190 || arena.timer == 60 || arena.timer == 30 || arena.timer == 10 || arena.timer <= 5) {
            ArenaHelper.sendMessage(arena, MessageKey.NORMAL_IN_GAME_ARENA_END, "1-" + arena.timer);
        }
        // Play a sound for the last 5 seconds of the arena game's duration
        if (arena.timer <= 5) {
            Location location = arena.arenaLocations.getLobbyWarp();
            if (location != null && location.getWorld() != null) {
                int pitch = (int) (2 - Math.floor(arena.timer / 2.0));
                location.getWorld().playSound(arena.arenaLocations.getLobbyWarp(),
                        Sound.ENTITY_EXPERIENCE_ORB_PICKUP, 1, pitch);
            }
        }

        for (Player player : arena.hiders) {
            updateHiddenState(player);
        }
        ScoreboardHelper.updateScoreboard(arena); // TODO Only do this when needed (player added/removed)
    }

    /**
     * Updates the hidden state of the given hider
     *
     * @param player <p>A player that's currently a hider</p>
     */
    private void updateHiddenState(Player player) {
        Location playerLocation = player.getLocation();
        Location moveLocation = MemoryStorage.moveLocation.get(player);

        // If the hider has somehow lost their disguise block from their inventory, 
        ItemStack inventoryBlock = player.getInventory().getItem(8);
        if (inventoryBlock == null && MemoryStorage.playerBlock.get(player) != null) {
            inventoryBlock = MemoryStorage.playerBlock.get(player);
            player.getInventory().setItem(8, inventoryBlock);
        }

        if (moveLocation == null || inventoryBlock == null) {
            return;
        }

        // If the player is standing still, try to turn the player into a solid block
        if (moveLocation.getX() == playerLocation.getX() && moveLocation.getY() == playerLocation.getY() &&
                moveLocation.getZ() == playerLocation.getZ()) {
            if (inventoryBlock.getAmount() > 1) {
                // Just decrease the timer
                inventoryBlock.setAmount(inventoryBlock.getAmount() - 1);
            } else {
                VisibilityHelper.makePlayerSolid(player, inventoryBlock);
            }
        } else {
            // The player moved. Un-hide and reset the hide timer
            inventoryBlock.setAmount(5);
            if (!DisguiseAPI.isDisguised(player)) {
                VisibilityHelper.makePlayerUnSolid(player);
            }
        }
    }

}
