package net.knarcraft.blockhunt;

import org.bukkit.GameMode;
import org.bukkit.Location;
import org.bukkit.inventory.ItemStack;
import org.bukkit.potion.PotionEffect;

import java.util.Collection;

/**
 * A record storing a player's data before they entered an arena
 *
 * @param playerLocation      <p>The location the player was at before joining the arena</p>
 * @param playerGameMode      <p>The game mode the player was in</p>
 * @param playerInventory     <p>The player's earlier inventory</p>
 * @param playerArmor         <p>The player's earlier equipped armor</p>
 * @param playerEXP           <p>The player's earlier exp amount</p>
 * @param playerEXPLevels     <p>The player's earlier exp level</p>
 * @param playerMaxHealth     <p>The player's previous max health</p>
 * @param playerHealth        <p>The player's previous health</p>
 * @param playerFood          <p>The player's previous food amount</p>
 * @param playerPotionEffects <p>The player's previous potion effects</p>
 * @param playerFlying        <p>The player's previous flying state</p>
 */
public record PlayerArenaData(Location playerLocation, GameMode playerGameMode, ItemStack[] playerInventory,
                              ItemStack[] playerArmor, Float playerEXP, Integer playerEXPLevels, Double playerMaxHealth,
                              Double playerHealth, Integer playerFood, Collection<PotionEffect> playerPotionEffects,
                              boolean playerFlying) {
}
