package net.knarcraft.blockhunt.command;

import net.knarcraft.blockhunt.config.MessageKey;
import net.knarcraft.blockhunt.manager.CommandManager;
import net.knarcraft.blockhunt.util.MessageHelper;
import net.knarcraft.blockhunt.util.TokensHelper;
import org.bukkit.Bukkit;
import org.bukkit.entity.Player;

public class TokensCommand extends DefaultCommand {


    @Override
    public boolean execute(Player player, String[] args) {
        if (args.length <= 3) {
            MessageHelper.sendMessage(player, MessageKey.ERROR_NOT_ENOUGH_ARGUMENTS, "syntax-" +
                    CommandManager.getTokensCommand().usage());
        } else {
            String option = args[1];
            String playerName = args[2];
            int amount;
            try {
                amount = Integer.parseInt(args[3]);
            } catch (NumberFormatException e) {
                MessageHelper.sendMessage(player, MessageKey.ERROR_NOT_A_NUMBER, "1-" + args[3]);
                return true;
            }

            Player tokenPlayer = Bukkit.getPlayer(playerName);
            if (tokenPlayer == null) {
                MessageHelper.sendMessage(player, MessageKey.ERROR_TOKENS_PLAYER_NOT_ONLINE, "%playerName%-" + playerName);
                return true;
            }
            String name = "\u00A78Console";
            if (player != null) {
                name = player.getName();
            }

            if (option.equalsIgnoreCase("set")) {
                updateTokens(player, tokenPlayer, name, amount, "Set", "to");
            } else if (option.equalsIgnoreCase("add")) {
                updateTokens(player, tokenPlayer, name, TokensHelper.getTokens(tokenPlayer) + amount,
                        "Added", "to");
            } else if (option.equalsIgnoreCase("take")) {
                updateTokens(player, tokenPlayer, name, TokensHelper.getTokens(tokenPlayer) - amount,
                        "Took", "from");
            } else {
                MessageHelper.sendMessage(player, MessageKey.ERROR_TOKENS_UNKNOWN_SETTING, "option-" + option);
            }
        }
        return true;
    }

    /**
     * Updates tokens for a player
     *
     * @param player      <p>The player that's changing tokens</p>
     * @param tokenPlayer <p>The player whose tokens are changed</p>
     * @param name        <p>The name of the player that's changing tokens</p>
     * @param newTokens   <p>The amount of tokens to set for the player</p>
     * @param action      <p>A description of the action performed</p>
     * @param pronoun     <p>The pronoun to use for the action. Ex: if the action is "Added", the correct pronoun is "to"</p>
     */
    private void updateTokens(Player player, Player tokenPlayer, String name, int newTokens, String action,
                              String pronoun) {
        TokensHelper.setTokens(player, newTokens);
        MessageHelper.sendMessage(player, MessageKey.NORMAL_TOKENS_CHANGED, "option-" + action,
                "%playerName%-" + tokenPlayer.getName(), "option2-" + pronoun, "amount-" + newTokens);
        MessageHelper.sendMessage(tokenPlayer, MessageKey.NORMAL_TOKENS_CHANGED_PERSON, "option-" +
                action.toLowerCase(), "%playerName%-" + name, "option2-" + pronoun, "amount-" + newTokens);
    }

}
