package net.knarcraft.blockhunt.listener;

import net.knarcraft.blockhunt.MemoryStorage;
import net.knarcraft.blockhunt.arena.Arena;
import net.knarcraft.blockhunt.arena.ArenaState;
import net.knarcraft.blockhunt.config.ConfigKey;
import net.knarcraft.blockhunt.config.MessageKey;
import net.knarcraft.blockhunt.config.Permission;
import net.knarcraft.blockhunt.manager.PermissionsManager;
import net.knarcraft.blockhunt.util.ArenaHelper;
import net.knarcraft.blockhunt.util.InventoryHelper;
import net.knarcraft.blockhunt.util.MessageHelper;
import net.knarcraft.blockhunt.util.SignsHelper;
import net.knarcraft.blockhunt.util.VisibilityHelper;
import net.md_5.bungee.api.ChatColor;
import org.bukkit.Bukkit;
import org.bukkit.Location;
import org.bukkit.Material;
import org.bukkit.Sound;
import org.bukkit.block.Block;
import org.bukkit.block.Sign;
import org.bukkit.block.sign.Side;
import org.bukkit.configuration.file.FileConfiguration;
import org.bukkit.entity.Player;
import org.bukkit.event.EventHandler;
import org.bukkit.event.EventPriority;
import org.bukkit.event.Listener;
import org.bukkit.event.block.Action;
import org.bukkit.event.player.PlayerInteractEvent;
import org.bukkit.inventory.Inventory;
import org.bukkit.inventory.ItemStack;
import org.bukkit.inventory.meta.ItemMeta;

import java.util.Map;

public class OnPlayerInteractEvent implements Listener {

    @EventHandler(priority = EventPriority.NORMAL, ignoreCancelled = true)
    public void onPlayerInteractEvent(PlayerInteractEvent event) {
        Player player = event.getPlayer();
        Block block = event.getClickedBlock();

        // Handle wand-usage
        if (PermissionsManager.hasPermission(player, Permission.CREATE, false)) {
            handleWandClick(event, player, block);
        }

        // Handle sign-clicking
        if (event.getAction() == Action.RIGHT_CLICK_BLOCK && block != null &&
                block.getState() instanceof Sign sign &&
                SignsHelper.isSign(block.getLocation())) {
            handleSignClick(player, sign);
            event.setCancelled(true);
            return;
        }

        // Block container or fire clicking
        if ((event.getAction() == Action.RIGHT_CLICK_BLOCK || event.getAction() == Action.LEFT_CLICK_BLOCK) &&
                block != null && !block.getType().isAir() &&
                (block.getType().isInteractable() ||
                        block.getRelative(event.getBlockFace()).getType().equals(Material.FIRE))) {
            if (MemoryStorage.playerArena.get(player) != null) {
                event.setCancelled(true);
            }
        }

        // Handle seekers hitting solid hiders
        if (event.getAction() == Action.LEFT_CLICK_BLOCK || event.getAction() == Action.LEFT_CLICK_BLOCK) {
            // TODO: Do this more efficiently and directly, for example by mapping location vector to player instead of
            //  the other way around.
            Arena arena = MemoryStorage.playerArena.get(event.getPlayer());
            if (arena != null && arena.seekers.contains(player)) {
                for (Player arenaPlayer : arena.playersInArena) {
                    // Check all hidden players
                    if (MemoryStorage.hiddenLocation.get(arenaPlayer) == null) {
                        continue;
                    }

                    // Check if the player is hidden at the location hit
                    Block moveLocBlock = MemoryStorage.hiddenLocation.get(arenaPlayer).getBlock();
                    if (block == null || moveLocBlock.getX() != block.getX() || moveLocBlock.getY() != block.getY() ||
                            moveLocBlock.getZ() != block.getZ()) {
                        continue;
                    }

                    // Make the hidden player no longer solid
                    MemoryStorage.moveLocation.put(arenaPlayer, new Location(arenaPlayer.getWorld(), 0, 0, 0));
                    arenaPlayer.getWorld().playSound(player.getLocation(), Sound.ENTITY_PLAYER_HURT, 1, 1);
                    VisibilityHelper.makePlayerUnSolid(arenaPlayer);
                }
            }
        }

        Arena arena = MemoryStorage.playerArena.get(player);
        if (arena == null) {
            return;
        }

        if (!arena.gameState.equals(ArenaState.WAITING) && !arena.gameState.equals(ArenaState.STARTING)) {
            return;
        }

        event.setCancelled(true);
        ItemStack item = player.getInventory().getItemInMainHand();
        if (item.getType() == Material.AIR || item.getItemMeta() == null) {
            return;
        }

        item.getItemMeta().getDisplayName();
        if (item.getItemMeta().getDisplayName().equals(MessageHelper.replaceAll(
                (String) MemoryStorage.config.get(ConfigKey.SHOP_BLOCK_CHOOSER_V_1_NAME)))) {
            Inventory blockChooser = Bukkit.createInventory(null, 36,
                    MessageHelper.replaceAll("\u00A7r" + MemoryStorage.config.get(
                            ConfigKey.SHOP_BLOCK_CHOOSER_V_1_NAME)));
            if (arena.disguiseBlocks != null) {
                for (int i = arena.disguiseBlocks.size(); i > 0; i = i - 1) {
                    blockChooser.setItem(i - 1, arena.disguiseBlocks.get(i - 1));
                }
            }

            player.openInventory(blockChooser);
        }

        if (item.getItemMeta().getDisplayName().equals(MessageHelper.replaceAll(
                (String) MemoryStorage.config.get(ConfigKey.SHOP_BLOCK_HUNT_PASS_V_2_NAME)))) {
            Inventory blockHuntPass = Bukkit.createInventory(null, 9,
                    MessageHelper.replaceAll("\u00A7r" + MemoryStorage.config.get(
                            ConfigKey.SHOP_BLOCK_HUNT_PASS_V_2_NAME)));
            ItemStack blockHuntPassSEEKER = new ItemStack(Material.BLUE_WOOL, 1);
            ItemMeta blockHuntPassItemMeta = blockHuntPassSEEKER.getItemMeta();
            blockHuntPassItemMeta.setDisplayName(MessageHelper.replaceAll("&eSEEKER"));
            blockHuntPassSEEKER.setItemMeta(blockHuntPassItemMeta);
            blockHuntPass.setItem(1, blockHuntPassSEEKER);

            ItemStack blockHuntPassHider = new ItemStack(Material.RED_WOOL, 1);
            blockHuntPassItemMeta.setDisplayName(MessageHelper.replaceAll("&eHIDER"));
            blockHuntPassHider.setItemMeta(blockHuntPassItemMeta);
            blockHuntPass.setItem(7, blockHuntPassHider);

            player.openInventory(blockHuntPass);
        }
    }

    private void handleSignClick(Player player, Sign sign) {
        FileConfiguration fileConfiguration = MemoryStorage.config.getFileConfiguration();
        String signIdentifier = sign.getSide(Side.FRONT).getLine(1);
        if (signIdentifier.equals(MessageHelper.replaceAll(
                fileConfiguration.getStringList(ConfigKey.SIGN_LEAVE.getPath()).get(1)))) {
            if (PermissionsManager.hasPermission(player, Permission.JOIN_SIGN, true)) {
                ArenaHelper.playerLeaveArena(player, true, true);
            }
        } else if (signIdentifier.equals(MessageHelper.replaceAll(
                fileConfiguration.getStringList(ConfigKey.SIGN_SHOP.getPath()).get(1)))) {
            if (PermissionsManager.hasPermission(player, Permission.SHOP, true)) {
                InventoryHelper.openShop(player);
            }
        } else {
            String cleanName = ChatColor.stripColor(ChatColor.translateAlternateColorCodes('&',
                    signIdentifier.toLowerCase()));
            Arena arena = MemoryStorage.arenaMap.get(cleanName);
            if (arena != null && PermissionsManager.hasPermission(player, Permission.JOIN_SIGN, true)) {
                ArenaHelper.playerJoinArena(player, arena.arenaName);
            }
        }
    }

    /**
     * Handles a player clicking a block with the selection wand
     *
     * @param event  <p>The triggered interaction event</p>
     * @param player <p>The player that interacted with a block</p>
     * @param block  <p>The block the player clicked</p>
     */
    private void handleWandClick(PlayerInteractEvent event, Player player, Block block) {
        ItemStack item = player.getInventory().getItemInMainHand();
        if (item.getType() == Material.AIR || item.getItemMeta() == null || !item.getItemMeta().hasDisplayName()) {
            return;
        }

        ItemMeta itemMeta = item.getItemMeta();
        if (!itemMeta.getDisplayName().equals(MessageHelper.replaceAll((String) MemoryStorage.config.get(ConfigKey.WAND_NAME)))) {
            return;
        }

        Action action = event.getAction();
        if (!event.hasBlock() || block == null) {
            return;
        }

        Location location = block.getLocation();
        if (!action.equals(Action.LEFT_CLICK_BLOCK) && !action.equals(Action.RIGHT_CLICK_BLOCK)) {
            return;
        }

        event.setCancelled(true);

        int number;
        Map<Player, Location> locationMap;
        if (action.equals(Action.LEFT_CLICK_BLOCK)) {
            locationMap = MemoryStorage.pos1;
            number = 1;
        } else {
            locationMap = MemoryStorage.pos2;
            number = 2;
        }

        if (locationMap.get(player) == null || !locationMap.get(player).equals(location)) {
            MessageHelper.sendMessage(player, MessageKey.NORMAL_WAND_SET_POSITION, "number-" + number,
                    "x-" + block.getX(), "y-" + block.getY(), "z-" + block.getZ());
            locationMap.put(player, location);
        }
    }

}
