package net.knarcraft.blockhunt.util;

import net.knarcraft.blockhunt.MemoryStorage;
import org.bukkit.entity.Player;

/**
 * A helper class for dealing with a player's tokens
 */
public final class TokensHelper {

    private TokensHelper() {

    }

    /**
     * Sets the amount of tokens for the given player
     *
     * @param player <p>The player to set tokens for</p>
     * @param tokens <p>The amount of tokens to set</p>
     */
    public static void setTokens(Player player, int tokens) {
        MemoryStorage.shop.getFileConfiguration().set(player.getName() + ".tokens", tokens);
        MemoryStorage.shop.save();
    }

    /**
     * Gets a player's current tokens
     *
     * @param player <p>The player to get tokens for</p>
     * @return <p>The player's current tokens</p>
     */
    public static int getTokens(Player player) {
        initializeTokens(player);
        return MemoryStorage.shop.getFileConfiguration().getInt(player.getName() + ".tokens");
    }

    /**
     * Makes sure the given player's tokens is initialized and not null
     *
     * @param player <p>The player whose tokens should be initialized</p>
     */
    private static void initializeTokens(Player player) {
        if (MemoryStorage.shop.getFileConfiguration().get(player.getName() + ".tokens") == null) {
            MemoryStorage.shop.getFileConfiguration().set(player.getName() + ".tokens", 0);
            MemoryStorage.shop.save();
        }
    }

}
